package main

import (
	"bufio"
	"bytes"
	"context"
	"encoding/json"
	"fmt"
	"github.com/joho/godotenv"
	"io"
	"log"
	"net/http"
	"os"
	"os/signal"
	"strings"
	"sync"
	"syscall"
)
import "time"

const wiegand_a = 17
const wiegand_b = 18
const wiegand_bit_timeout = time.Millisecond * 8
const solenoid = 21

type card struct {
	UidHash string `json:"uid_hash"`
}

type cardList struct {
	AllowedUids []struct {
		Token card `json:"token"`
	} `json:"allowed_uids"`
	KeepOpenUntil *time.Time `json:"keep_open_until,omitempty"`
}

type ValidUids map[string]bool // bool has no meaning

type Config struct {
	door         string
	uidSalt      string
	doorOpenTime string
	mock         bool
	api          struct {
		allowed  string
		longpoll string
		swipe    string
		key      string
	}
}

type KeepDoorOpen struct {
	until time.Time
	timer *time.Timer
}

var config Config
var globalLock sync.Mutex
var validUids ValidUids
var wiegand Wiegand
var keepDoorOpen KeepDoorOpen

func main() {

	ctx, cancel := signal.NotifyContext(context.Background(), os.Interrupt, syscall.SIGTERM)
	defer cancel()

	godotenv.Load()

	config.door = os.Getenv("KDOORPI_DOOR")
	config.api.allowed = os.Getenv("KDOORPI_API_ALLOWED")
	config.api.longpoll = os.Getenv("KDOORPI_API_LONGPOLL")
	config.api.swipe = os.Getenv("KDOORPI_API_SWIPE")
	config.api.key = os.Getenv("KDOORPI_API_KEY")
	config.uidSalt = os.Getenv("KDOORPI_UID_SALT")
	config.doorOpenTime = os.Getenv("KDOORPI_OPEN_TIME")
	_, config.mock = os.LookupEnv("KDOORPI_MOCK_HW")

	go func() {
		setup()
	}()

	<-ctx.Done()
	log.Printf("Cleanup\n")

	// cleanup
}
func setup() {
	log.Println("Started Setup")

	if config.mock {
		log.Println("MOCK mode enabled")
		if config.door == "" {
			config.door = "mockdoor"
		}
		wiegand = &WiegandMock{}
	} else {
		wiegand = WiegandSetup(wiegand_a, wiegand_b, wiegand_bit_timeout, solenoid)
	}
	log.Println("HW Setup done")

	go func() {
		for {
			waitEvents()
		}
	}()

	log.Println("Initialized longpoll event loop")

	http.DefaultClient.Timeout = 120 * time.Second

	for {
		log.Println("Start initial token population")
		err := reloadTokens()
		if err == nil {
			break
		}
		log.Printf("Initial token population failed. err: %v", err)
		log.Println("Retrying in 10 seconds...")
		time.Sleep(10 * time.Second)
	}

	log.Println("Initial token population success")

	go cardRunner(wiegand)

	log.Println("Setup completed")
}

func OpenAndCloseDoor(w Wiegand) error {
	err := w.OpenDoor()
	if err != nil {
		return err
	}

	if keepDoorOpen.until.After(time.Now().Add(5 * time.Second)) {
		fmt.Println("Door is already open")
		return nil
	}

	fmt.Println("Door is now open")

	time.Sleep(5 * time.Second)

	err = w.CloseDoor()
	if err != nil {
		return err
	}
	fmt.Println("Door is now closed")

	return nil
}

func cardRunner(w Wiegand) {
	for {
		card, err := w.GetCardUid()
		if err != nil {
			continue
		}

		printCardId(card)
		hashedHex := hashCardUid(card)
		log.Println(hashedHex)

		globalLock.Lock()
		ok := validUids[hashedHex]
		globalLock.Unlock()

		go func() {
			err := sendSwipeEvent(hashedHex, ok)
			if err != nil {
				log.Println("Failed to send swipe event: %v", err)
			}
		}()

		if ok {
			log.Println("Opening door")
			err := OpenAndCloseDoor(w)
			if err != nil {
				log.Println("There was an error opening and closing the Door")
			}
		} else {
			log.Println("Unknown card")
		}

	}
}

func ParseNextMessage(r *bufio.Reader) (string, error) {
	var message string

	for {
		s, err := r.ReadString('\n')
		if err != nil {
			return message, err
		}

		message = message + s
		nextBytes, err := r.ReadByte()
		if err != nil {
			return message, err
		}

		if nextBytes == '\n' {
			return message, nil
		}

		r.UnreadByte()
	}

}

func waitEvents() error {
	req, err := http.NewRequest(http.MethodGet, config.api.longpoll, nil)
	if err != nil {
		return err
	}
	req.Header.Add("KEY", config.api.key)
	resp, err := http.DefaultClient.Do(req)
	if err != nil {
		return err
	}
	log.Printf("%v\n", resp)

	reader := bufio.NewReader(resp.Body)
	for {
		msg, err := ParseNextMessage(reader)
		if err != nil {
			return err
		}

		for _, line := range strings.Split(msg, "\n") {
			data, found_data := strings.CutPrefix(line, "data:")
			if !found_data {
				continue
			}
			log.Printf("got server data: %q\n", data)
			if strings.TrimSpace(data) == config.door {
				err := OpenAndCloseDoor(wiegand)
				if err != nil {
					log.Println("There was an error opening and closing the Door")
				}
			}

			go reloadTokens()
		}

	}

	log.Printf("%v\n", resp)
	return nil
}

func reloadTokens() error {
	req, err := http.NewRequest(http.MethodGet, config.api.allowed, nil)
	if err != nil {
		return err
	}
	req.Header.Add("KEY", config.api.key)
	resp, err := http.DefaultClient.Do(req)
	if err != nil {
		return err
	}
	log.Printf("%v\n", resp)

	var cl cardList

	body, err := io.ReadAll(resp.Body)
	if err != nil {
		return err
	}

	err = json.Unmarshal(body, &cl)
	if err != nil {
		return err
	}

	globalLock.Lock()
	defer globalLock.Unlock()
	validUids = make(ValidUids)
	for i, val := range cl.AllowedUids {
		log.Printf("%d: %+v\n", i, val.Token.UidHash)
		validUids[val.Token.UidHash] = true
	}

	if cl.KeepOpenUntil != nil {
		updateKeepOpenDoor(*cl.KeepOpenUntil)
	}

	return nil
}

func updateKeepOpenDoor(newKeepOpenTime time.Time) {

	// is there one active?
	if keepDoorOpen.timer != nil {
		keepDoorOpen.timer.Stop()
		keepDoorOpen = KeepDoorOpen{}
	}

	if newKeepOpenTime.After(time.Now()) {
		log.Printf("Keeping door open until %v", newKeepOpenTime)
		wiegand.OpenDoor()
		timer := time.AfterFunc(time.Until(newKeepOpenTime), handleKeepDoorOpenCloseCleanup)
		keepDoorOpen = KeepDoorOpen{
			timer: timer,
			until: newKeepOpenTime,
		}
	} else {
		wiegand.CloseDoor()
	}
}

func handleKeepDoorOpenCloseCleanup() {
	fmt.Println("Keep door open time is reached!")
	wiegand.CloseDoor()
	keepDoorOpen = KeepDoorOpen{}
}

func sendSwipeEvent(cardUidHash string, success bool) error {
	swipeEvent := map[string]string{
		"uid_hash":  cardUidHash,
		"door":      config.door,
		"timestamp": time.Now().Format(time.DateTime),
		"success":   fmt.Sprint(success),
	}

	data, err := json.Marshal(swipeEvent)
	if err != nil {
		return err
	}

	req, err := http.NewRequest(http.MethodPost, config.api.swipe, bytes.NewReader(data))
	if err != nil {
		return err
	}
	req.Header.Add("KEY", config.api.key)
	resp, err := http.DefaultClient.Do(req)
	if err != nil {
		return err
	}

	if !(resp.StatusCode >= 200 && resp.StatusCode < 300) {
		return fmt.Errorf("Server responded with %d", resp.StatusCode)
	}

	return nil
}
